from queue import Empty, Full, Queue
from typing import Any, NamedTuple, Optional

from result import Err, Ok, Result

from ..util import _get_gst
from .converter import ConverterBase, ConverterRaw
from .exception import PipelineBuildError
from .pipeline import PipelineGenerator, _BuiltPipeline

__all__ = [
    "GstStreamBuilder",
]


class GstStreamBuilder:
    _pipeline_generator: PipelineGenerator
    _converter: ConverterBase

    def __init__(self, pipeline_generator: PipelineGenerator, converter: Optional[ConverterBase] = None):
        """
        args:
            - pipeline_generator: :class:`~PipelineGenerator`
            - converter: :class:`~ConverterBase`, defaults to :class:`~ConverterRaw`.
        """

        if converter is None:
            converter = ConverterRaw()

        assert isinstance(
            pipeline_generator, PipelineGenerator
        ), f"pipeline_generator should be instance of PipelineGenerator, but got: {type(pipeline_generator)}"
        assert isinstance(
            converter, ConverterBase
        ), f"converter should be instance of ConverterBase, but got: {type(converter)}"

        self._pipeline_generator = pipeline_generator
        self._converter = converter

    def start_streaming(self) -> "_GstStream":  # noqa F821 (Hey linter, see below.)
        """
        return:
            - :class:`~_GstStream`
        exceptions:
            - :class:`~PipelineBuildError`
        """

        built_pipeline_ = self._pipeline_generator.build()
        if built_pipeline_.is_err():
            raise built_pipeline_.unwrap_err()
        built_pipeline = built_pipeline_.unwrap()
        inner = Inner(built_pipeline, self._converter)
        return _GstStream(inner)


class _GstStream:
    _inner: "Inner"  # noqa F821 (Hey linter, see below.)

    def __init__(self, inner: "Inner"):  # noqa F821 (Hey linter, see below.)
        self._inner = inner

    def __enter__(self) -> "_GstStream":  # noqa F821 (Hey linter, see above.)
        err = self._inner.start()
        if err.is_err():
            raise err.unwrap_err()

        return self

    def __exit__(self, _ex_type: Any, _ex_value: Any, _trace: Any) -> bool:  # type: ignore
        # Forgot errors in stopping pipeline.
        _err = self._inner.stop()  # noqa F841

        return False

    def is_running(self) -> bool:
        return self._inner.is_running()

    # Here, Any = ConverterBase::ConvertResult, but we can't yet express associated types.
    # c.f. https://github.com/python/mypy/issues/7790
    def capture(self, timeout_secs: float) -> Any:
        res = self._inner.capture(timeout_secs)
        if res.is_ok():
            return res.unwrap()
        else:
            raise res.unwrap_err()


class InternalMessageKind:
    FROM_NEW_SAMPLE = 0
    FROM_MESSAGE = 1


class InternalMessage(NamedTuple):
    kind: int  # InternalMessageKind
    payload: Any


class Inner:
    _Gst: "Gst"  # type: ignore  # noqa F821
    _built_pipeline: _BuiltPipeline
    _converter: ConverterBase
    _queue: "Queue[InternalMessage]"
    _is_running: bool
    _bus: "Gst.Bus"  # type: ignore  # noqa F821

    def __init__(self, built_pipeline: _BuiltPipeline, converter: ConverterBase):
        self._Gst = _get_gst()
        self._built_pipeline = built_pipeline
        self._converter = converter
        self._queue = Queue(1)
        self._is_running = False

        self._built_pipeline.sink.connect("new-sample", self._cb_new_sample)
        self._bus = self._built_pipeline.pipeline.get_bus()
        self._bus.add_signal_watch()
        self._bus.connect("message::eos", self._cb_message)
        self._bus.connect("message::error", self._cb_message)

    def is_running(self) -> bool:
        return self._is_running

    def _change_pipeline_state(
        self,
        desired: "Gst.State",  # type: ignore  # noqa F821
    ) -> Result[None, PipelineBuildError]:
        """
        Blocking function to change pipeline state to be `desired` or fail.

        args:
            - desired: One of enum `~Gst.State`.
        """

        self._built_pipeline.pipeline.set_state(desired)
        # Note that x is _ResultTuple of type (<Gst.StateChangeReturn>, state=<Gst.State>, pending=<Gst.State>).
        x = self._built_pipeline.pipeline.get_state(self._Gst.CLOCK_TIME_NONE)
        if x[0] == self._Gst.StateChangeReturn.FAILURE:
            return Err(PipelineBuildError(f"failed to change state of pipeline: desired = {desired}, {x}"))
        elif x.state == desired:
            return Ok(None)
        else:
            raise RuntimeError("unreachable")

    def start(self) -> Result[None, PipelineBuildError]:
        res = self._change_pipeline_state(self._Gst.State.PLAYING)
        if res.is_err():
            return res

        self._is_running = True
        return Ok(None)

    def stop(self) -> Result[None, PipelineBuildError]:
        if self._is_running:
            self._is_running = False
            self._bus.remove_signal_watch()
            return self._change_pipeline_state(self._Gst.State.NULL)
        else:
            return Ok(None)

    def capture(self, timeout_secs: float) -> Result[Optional[Any], Exception]:
        im: Optional[InternalMessage]
        try:
            im = self._queue.get(block=True, timeout=timeout_secs)
        except Empty:
            im = None

        if im is None:
            return Ok(None)
        elif im.kind == InternalMessageKind.FROM_NEW_SAMPLE:
            # Note that there is a case we cannot get sample via `pull-sample` while got `new-sample` signal:
            # (This is because we decoupled these signals.)
            #
            #   1. (thread A) Called this method `capture`.
            #   2. (thread B) Got a signal `new-sample`.  Enqueue.
            #   3. (thread A) Dequeue & interrupted.
            #   4. (thread B) Got a signal `new-sample`.  Enqueue.
            #   5. (thread A) Emit `pull-sample` and process.
            #   6. (thread A) Called `capture`, dequeue, emit, but there is no sample if `max-buffers=1`.
            #
            # While lots of examples (e.g., https://gstreamer.freedesktop.org/documentation/tutorials/basic/short-cutting-the-pipeline.html)
            # emit `pull-sample` in `new-sample` callback, we use this decoupling because this affects performance in python case.
            sample = self._built_pipeline.sink.emit("pull-sample")
            if sample is None:
                return Ok(None)
            else:
                return self._converter.convert_sample(sample)
        elif im.kind == InternalMessageKind.FROM_MESSAGE:
            message = im.payload
            if message.type == self._Gst.MessageType.EOS:
                self.stop()
                return Ok(None)
            elif message.type == self._Gst.MessageType.ERROR:
                return Err(Exception(message))
            else:
                raise RuntimeError("unreachable")
        else:
            raise RuntimeError("unreachable")

    def _cb_new_sample(self, _: Any) -> "Gst.FlowReturn":  # type: ignore  # noqa F821
        im = InternalMessage(InternalMessageKind.FROM_NEW_SAMPLE, None)
        try:
            self._queue.put_nowait(im)
        except Full:
            pass
        return self._Gst.FlowReturn.OK

    def _cb_message(self, _: Any, message: Any):  # type: ignore
        im = InternalMessage(InternalMessageKind.FROM_MESSAGE, message)
        self._queue.put(im)


# For debug.
# class _DummyMessage:
#     def __init__(self, t: Any):
#         self.type = t



import logging as _logging

# noqa idiom
if True:
    logger = _logging.getLogger(__name__)
    logger.addHandler(_logging.NullHandler())

import copy
from typing import Any, Dict

from .pipeline import AppsinkColorFormat, PipelineBuilder, PipelineGenerator

__all__ = [
    "videotestsrc",
    "rtsp_h264",
]


DEFAULT_CAPS = {
    "width": 640,
    "height": 480,
}


def videotestsrc(pattern: str = "smpte", caps: Dict[str, Any] = DEFAULT_CAPS) -> PipelineGenerator:
    """
    Create a pipeline like:
        videotestsrc pattern=<pattern>
        ! video/x-raw,format=RGB,... \
        ! appsink

    args:
        - pattern: `str`, defaults "smpte".
        - caps: `dict`,
            {
                'width': int,
                'height': int,
                'framerate': Option[int], // Default: 10
            }
    returns:
        - :class:`~PipelineGenerator`
    """

    caps = copy.copy(caps)
    assert "width" in caps
    assert "height" in caps
    if "framerate" not in caps:
        caps["framerate"] = 10

    return (
        PipelineBuilder(force_format=AppsinkColorFormat.RGB)
        .add(
            "videotestsrc",
            {"pattern": pattern},
        )
        .add("videoscale")
        .add_appsink_with_caps(
            {
                "max-buffers": 1,
                "drop": True,
                "emit-signals": True,
            },
            caps,
        )
        .finalize()
    )


def rtsp_h264(
    proxy: str,
    location: str,
    protocols: str,
    decoder_type: str,
    caps: Dict[str, Any] = DEFAULT_CAPS,
) -> PipelineGenerator:
    """
    Create a pipeline like:
        rtspsrc proxy=<proxy> location=<location> \
        ! rtph264depay ! h264parse ! <decoder> \
        ! videorate ! videoscale ! videoconvert \
        ! video/x-raw,format=RGB,... \
        ! appsink
    where
        <decoder> = v4l2h264dec (if decoder_type == 'v4l2')
                  = omxh264dec (if decoder_type == 'omx')

    args:
        - proxy: proxy URL 'tcp://...'
        - location: rtsp resource location URL 'rtsp://<host>:<port>/<path>'
        - decoder_type: string, 'v4l2' | 'omx' | 'libav'
        - caps: `dict`
            {
                'width': int,
                'height': int,
                'framerate': Option[int], // Default: 10
            }
    returns:
        - :class:`~PipelineGenerator`
    """

    if decoder_type == "v4l2":
        decoder = "v4l2h264dec"
    elif decoder_type == "omx":
        decoder = "omxh264dec"
    elif decoder_type == "libav":
        decoder = "avdec_h264"
    else:
        raise ValueError(f"decoder_type should be 'v4l2' | 'omx' | 'libav', but got: {decoder_type}")

    return _rtsp_h264(proxy, location, protocols, decoder, caps)


def _rtsp_h264(
    proxy: str,
    location: str,
    protocols: str,
    decoder: str,
    caps: Dict[str, Any],
) -> PipelineGenerator:
    assert "width" in caps
    assert "height" in caps
    assert "framerate" in caps

    rtspsrc_props = {
        "location": location,
        "protocols": protocols,
        "latency": 0,
        "max-rtcp-rtp-time-diff": 100,
        "drop-on-latency": True,
    }
    if proxy is not None:
        rtspsrc_props["proxy"] = proxy

    return (
        PipelineBuilder(force_format=AppsinkColorFormat.RGB)
        .add("rtspsrc", rtspsrc_props)
        .add("rtph264depay")
        .add("h264parse")
        .add(decoder)
        .add(
            "videorate",
            {
                # We don't use `drop-only` because omxh264dec generates a frame with `framerate=0/1`
                # in the startup for some cameras, and it causes a fatal error.
                # 'drop-only': True,
                "skip-to-first": True,
            },
        )
        .add("videoscale")
        .add("videoconvert")
        .add_appsink_with_caps(
            {
                "max-buffers": 1,
                "drop": True,
                "emit-signals": True,
            },
            caps,
        )
        .finalize()
    )



from .gstreamer.exception import GstNotInitializedError

__all__ = []  # type: ignore


CACHED_GST = None


def _get_gst() -> "Gst":  # type: ignore  # noqa F821
    global CACHED_GST

    if CACHED_GST is None:
        try:
            import gi

            gi.require_version("Gst", "1.0")
            from gi.repository import Gst

            CACHED_GST = Gst
            if not Gst.is_initialized():
                raise GstNotInitializedError()
        except Exception as e:
            raise GstNotInitializedError() from e

    return CACHED_GST



import logging as _logging

# noqa idiom
if True:
    logger = _logging.getLogger(__name__)
    logger.addHandler(_logging.NullHandler())

import time
from typing import Optional

from actfw_core.task import Producer
from PIL.Image import Image as PIL_Image

from .gstreamer.exception import ConnectionLostError, PipelineBuildError
from .gstreamer.stream import GstStreamBuilder
from .restart_handler import Restart, RestartHandlerBase, Stop

__all__ = [
    "GstreamerCapture",
]


class GstreamerCapture(Producer[PIL_Image]):
    _builder: GstStreamBuilder
    _restart_handler: RestartHandlerBase

    def __init__(self, builder: GstStreamBuilder, restart_handler: RestartHandlerBase):
        """
        Captured Frame Producer using GStreamer.

        Genrates `Frame`s of given `GonverterBase::ConvertResult`, where `ConverterBase` is in `builder`.

        args:
            - builder: :class:`~GstStreamBuilder`
            - restart_handler: :class:`~RestartHandlerBase`
        """

        assert isinstance(
            builder, GstStreamBuilder
        ), f"builder should be instance of GstStreamBuilder, but got: {type(builder)}"
        assert isinstance(
            restart_handler, RestartHandlerBase
        ), f"restart_handler should be instance of RestartHandler, but got: {type(restart_handler)}"

        super().__init__()

        self._builder = builder
        self._restart_handler = restart_handler

    def run(self) -> None:
        connection_lost_threshold = self._restart_handler.connection_lost_secs_threshold()

        try:
            while True:
                try:
                    self._loop(connection_lost_threshold)
                except PipelineBuildError as e:
                    logger.debug(e)

                    action = self._restart_handler.pipeline_build_error(e)
                    if isinstance(action, Stop):
                        return None
                    elif isinstance(action, Restart):
                        continue
                    else:
                        raise RuntimeError("unreachable")
                except ConnectionLostError as e:
                    logger.debug(e)

                    action = self._restart_handler.connection_lost(e)
                    if isinstance(action, Stop):
                        return None
                    elif isinstance(action, Restart):
                        continue
                    else:
                        raise RuntimeError("unreachable")

                break
        finally:
            self.stop()

    def _loop(self, connection_lost_threshold: Optional[float]) -> None:
        no_sample_start: Optional[float] = None
        with self._builder.start_streaming() as stream:
            while self._is_running():
                if not stream.is_running():
                    raise ConnectionLostError()

                if (connection_lost_threshold is not None) and (no_sample_start is not None):
                    if (time.time() - no_sample_start) > connection_lost_threshold:
                        raise ConnectionLostError()

                value = stream.capture(timeout_secs=1)
                if value is None:
                    if no_sample_start is None:
                        no_sample_start = time.time()
                else:
                    no_sample_start = None
                    self._outlet(value)



from abc import ABC, abstractmethod
from typing import Optional, Union

from .gstreamer.exception import ConnectionLostError, PipelineBuildError

__all__ = [
    "RestartAction",
    "Stop",
    "Restart",
    "RestartHandlerBase",
    "SimpleRestartHandler",
]


class Stop:
    pass


class Restart:
    pass


RestartAction = Union[Stop, Restart]


class RestartHandlerBase(ABC):
    @abstractmethod
    def connection_lost_secs_threshold(self) -> Optional[float]:
        """
        :class:`~GstCapture.run` waits a new frame this seconds.  If cannot get no frames more than
        this seconds, raise :class:`~ConnectionLostError`.
        """

        raise NotImplementedError()

    @abstractmethod
    def pipeline_build_error(self, err: PipelineBuildError) -> RestartAction:
        """
        Called from :class:`~GstCapture.run` when got :class:`~PipelineBuildError`.
        """

        raise NotImplementedError()

    @abstractmethod
    def connection_lost(self, err: ConnectionLostError) -> RestartAction:
        """
        Called from :class:`~GstCapture.run` when got :class:`~ConnectionLostError`.
        """

        raise NotImplementedError()


class SimpleRestartHandler(RestartHandlerBase):
    _connection_lost_secs_threshold: int
    _error_count_threshold: int
    _error_count: int

    def __init__(self, connection_lost_secs_threshold: int, error_count_threshould: int) -> None:
        self._connection_lost_secs_threshold = connection_lost_secs_threshold
        self._error_count_threshold = error_count_threshould
        self._error_count = 0

    def connection_lost_secs_threshold(self) -> Optional[float]:
        return self._connection_lost_secs_threshold

    def pipeline_build_error(self, err: PipelineBuildError) -> RestartAction:
        self._error_count += 1

        if self._error_count_threshold < self._error_count:
            raise err
        else:
            print("pipeline build error.  restarting...", flush=True)

            return Restart()

    def connection_lost(self, err: ConnectionLostError) -> RestartAction:
        self._error_count += 1

        if self._error_count_threshold < self._error_count:
            raise err
        else:
            print("suspicious connection lost.  restarting...", flush=True)

            return Restart()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['actfw_gstreamer', 'actfw_gstreamer.gstreamer']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=5,<6', 'PyGObject>=3,<4', 'actfw-core==2.0.0', 'result>=0.6.0,<0.7.0']

setup_kwargs = {
    'name': 'actfw-gstreamer',
    'version': '0.1.0a8',
    'description': "actfw's additional components using GStreamer",
    'long_description': '# actfw-gstreamer\n\nactfw\'s components using GStreamer for implementation.\nactfw is a framework for Actcast Application written in Python.\n\n## Installation\n\n```console\nsudo apt-get update\nsudo apt-get install -y python3-pip python3-pil \nsudo apt-get install libgstreamer1.0-dev libgirepository1.0-dev ibgstreamer-plugins-base1.0-dev libglib2.0-dev\npip3 install actfw-gstreamer\n```\n\n## Document\n\n- [API References](https://idein.github.io/actfw-gstreamer/latest/)\n\n## Usage\n\nSee [actfw-core](https://github.com/Idein/actfw-core) for basic usage of `actfw` framework.\n\n### Initalization\n\nAn application using `actfw-gstreamer` have to initialize GStreamer library before using `actfw-gstreamer`\'s components.\n\n```python\nif __name__ == \'__main__\':\n    import gi\n\n    gi.require_version(\'Gst\', \'1.0\')\n    from gi.repository import Gst\n    Gst.init(None)\n\n    main()\n```\n\n### `videotestsrc`\n\nYou can learn basic usage of `actfw-gstreamer` by using `videotestsrc`.\n\n```python\nfrom actfw_gstreamer.capture import GstreamerCapture\nfrom actfw_gstreamer.gstreamer.converter import ConverterPIL\nfrom actfw_gstreamer.gstreamer.stream import GstStreamBuilder\nfrom actfw_gstreamer.restart_handler import SimpleRestartHandler\n\n\ndef videotestsrc_capture() -> GstreamerCapture:\n    pipeline_generator = preconfigured_pipeline.videotestsrc()\n    builder = GstStreamBuilder(pipeline_generator, ConverterPIL())\n    restart_handler = SimpleRestartHandler(10, 5)\n    return GstreamerCapture(builder, restart_handler)\n\n\ndef main():\n    app = actfw_core.Application()\n\n    capture = videotestsrc_capture()\n    app.register_task(capture)\n\n    consumer = YourConsumer()\n    app.register_task(consumer)\n\n    capture.connect(consumer)\n\n    app.run()\n```\n\nThis generates [`Frame`](https://idein.github.io/actfw-core/latest/actfw_core.html#actfw_core.capture.Frame)s using [videotestsrc](https://gstreamer.freedesktop.org/documentation/videotestsrc/index.html).\n\n- `GstreamerCapture` is a [`Producer`](https://idein.github.io/actfw-core/latest/actfw_core.task.html#actfw_core.task.producer.Producer).\n  - It generates `Frame`s consists of an output of `ConverterBase`.  In this case, converter class is `ConverterPIL` and output is `PIL.Image.Image`.\n- `GstStreamBuilder` and `PipelineGenerator` determines how to build gstreamer pipelines.\n- `preconfigured_pipeline` provides preconfigured `PipelineGenerator`s.\n- `SimpleRestartHandler` is a simple implementation of `RestartHandlerBase`, which determines "restart strategy".\n\nFor more details, see [tests](tests/intergation_test/test_gstreamer_output.py).\n\n### `rtspsrc`\n\nYou can use [rtspsrc](https://gstreamer.freedesktop.org/documentation/rtsp/rtspsrc.html) using `preconfigured_pipeline.rtsp_h264()`.\n\nNote that, as of now (2021-04), [Actcast application](https://actcast.io/docs/ForVendor/ApplicationDevelopment/) cannot use multicast UDP with dynamic address and unicast UDP.\n(RTSP client communicates with RTSP server in RTP and determines adderss of mulitcast UDP.)\nTherefore, you can use only the option `protocols = "tcp"`.\nSee also https://gstreamer.freedesktop.org/documentation/rtsp/rtspsrc.html#rtspsrc:protocols .\n\nYou should also pay attention to decoders. Available decoders are below:\n\n| decoder (package) \\ device                                     | Raspberry Pi 3 | Raspberry Pi 4 | Jetson Nano |\n| -------------------------------------------------------------- | -------------- | -------------- | ----------- |\n| `omxh264` (from `gstreamer1.0-omx` and `gstreamer1.0-omx-rpi`) | o              | x              | ?           |\n| `v4l2h264dec` (from `gstreamer1.0-plugins-good`)               | very slow      | o              | ?           |\n\nIf your application supports various devices, you should branch by hardware types and select appropriate `decoder_type`.\nFor example, it is recommended to use `decoder_type` `omx` for Raspberry Pi 3 and `v4l2` for Raspberry Pi 4.\nCurrently, this library does not provide auto determination.\n\n## Development Guide\n\n### Installation of dev requirements\n\n```console\npip3 install poetry\npoetry install\n```\n\n### Running tests\n\n```console\npoetry run nose2 -v\n```\n\n### Releasing package & API doc\n\nCI will automatically do.\nFollow the following branch/tag rules.\n\n1. Make changes for next version in `master` branch (via pull-requests).\n2. Make a PR that updates version in `pyproject.toml` and merge it to `master` branch.\n3. Create Git tag from `master` branch\'s HEAD named `release-<New version>`. E.g. `release-1.4.0`.\n4. Then CI will build/upload package to PyPI & API doc to GitHub Pages.\n',
    'author': 'Idein Inc.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Idein/actfw-gstreamer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



__all__ = [
    "GstNotInitializedError",
    "PipelineBuildError",
    "ConnectionLostError",
]


class GstNotInitializedError(RuntimeError):
    def __init__(self) -> None:
        super().__init__("You must initialize Gst first. See document.")


class PipelineBuildError(RuntimeError):
    pass


class ConnectionLostError(RuntimeError):
    pass



import logging as _logging

# noqa idiom
if True:
    logger = _logging.getLogger(__name__)
    logger.addHandler(_logging.NullHandler())

import enum
from typing import Any, Dict, List, NamedTuple, Optional

from result import Err, Ok, Result

from ..util import _get_gst
from .exception import PipelineBuildError

__all__ = [
    "AppsinkColorFormat",
    "PipelineBuilder",
    "PipelineGenerator",
]


class AppsinkColorFormat(enum.Enum):
    BGR = enum.auto()
    RGB = enum.auto()
    RGBx = enum.auto()

    @classmethod
    def _from_caps_format(cls, format_: str) -> Result["AppsinkColorFormat", ValueError]:
        if format_ in cls._FROM_CAPS_FORMAT:  # type: ignore
            return Ok(cls._FROM_CAPS_FORMAT[format_])  # type: ignore
        else:
            return Err(ValueError(f"Unknown format: {format_}"))

    def _to_caps_format(self) -> str:
        return self._TO_CAPS_FORMAT[self]  # type: ignore

    def _to_PIL_raw_mode(self) -> str:
        return self._TO_PIL_RAW_MODE[self]  # type: ignore


# I know metaclass trick, but it's enough.
_CORR = {
    AppsinkColorFormat.BGR: ("BGR", "BGR"),
    AppsinkColorFormat.RGB: ("RGB", "RGB"),
    AppsinkColorFormat.RGBx: ("RGBx", "RGBX"),
}
AppsinkColorFormat._FROM_CAPS_FORMAT = (  # type: ignore
    # fmt: off
    dict((caps_format, x)
         for (x, (caps_format, _)) in _CORR.items())
)
AppsinkColorFormat._TO_CAPS_FORMAT = (  # type: ignore
    # fmt: off
    dict((x, caps_format)
         for (x, (caps_format, _)) in _CORR.items())
)
AppsinkColorFormat._TO_PIL_RAW_MODE = (  # type: ignore
    # fmt: off
    dict((x, raw_mode)
         for (x, (_, raw_mode)) in _CORR.items())
)


def _make_element(Gst: "Gst", element: str, props: Dict[str, Any]) -> "Gst.Element":  # type: ignore  # noqa F821
    """
    exceptions:
        - :class:`~PipelineBuildError`
    """

    x = Gst.ElementFactory.make(element)
    if not x:
        raise PipelineBuildError(f"failed to create `{element}`")

    for (k, v) in props.items():
        logger.info(f"set property: {k} => {v}")
        x.set_property(k, v)

    return x


def _make_capsfilter(Gst: "Gst", caps_string: str) -> "Gst.Element":  # type: ignore  # noqa F821
    """
    exceptions:
        - :class:`~PipelineBuildError`
    """

    caps = Gst.caps_from_string(caps_string)
    return _make_element(Gst, "capsfilter", {"caps": caps})


class PipelineBuilder:
    _Gst: "Gst"  # type: ignore  # noqa F821
    _thunks: List[Any]
    _caps_string: Optional[str]
    _finalized: bool

    def __init__(self, force_format: Optional[AppsinkColorFormat] = None):
        self._Gst = _get_gst()
        self._thunks = []
        self._caps_string = None
        self._finalized = False

        if force_format is None:
            self._caps_base = "video/x-raw,format=(string){BGR,RGB,RGBx}"
        else:
            format_ = force_format._to_caps_format()
            self._caps_base = f"video/x-raw,format={format_}"

    def is_finalized(self) -> bool:
        return self._finalized

    def add(self, element: str, props: Dict[str, Any] = {}) -> "PipelineBuilder":  # noqa B006
        self._thunks.append(lambda: _make_element(self._Gst, element, props))
        return self

    def add_capsfilter(self, caps_string: str) -> "PipelineBuilder":  # noqa F821
        self._thunks.append(lambda: _make_capsfilter(self._Gst, caps_string))
        return self

    def add_appsink_with_caps(self, props: Dict[str, Any] = {}, caps: Dict[str, Any] = {}) -> "PipelineBuilder":  # noqa B006
        """
        Effect: Change `self.is_finalized()` to be true.

        args:
            - caps: `dict`
                {
                    'width': int,
                    'height': int,
                    'framerate': Optional[int], # Used as `framerate={framerate}/1`.
                }
        """

        assert "width" in caps
        assert "height" in caps
        assert "framerate" in caps

        self.add("appsink", props)

        s = self._caps_base
        for key in ["width", "height"]:
            if key in caps:
                s += f",{key}={caps[key]}"
        framerate = caps["framerate"]
        if framerate is not None:
            s += f",framerate={framerate}/1"
        self._caps_string = s
        self._finalized = True

        return self

    def finalize(self) -> "PipelineGenerator":  # noqa F821 (Hey linter, see below.)
        """
        returns:
            - :class:`~PipelineGenerator`
        """

        assert self._finalized

        return PipelineGenerator(self._thunks, self._caps_string)  # type: ignore


class PipelineGenerator:
    """
    Users should make instances of this class through :class:`~PipelineBuilder`.
    """

    _Gst: "Gst"  # type: ignore  # noqa F821
    _thunks: List[Any]
    _caps_string: Optional[str]

    def __init__(self, thunks: List[Any], caps_string: str):
        # It definitely contains `appsink`.
        assert len(thunks) > 0

        self._Gst = _get_gst()
        self._thunks = thunks
        self._caps_string = caps_string

    def build(self) -> Result["_BuiltPipeline", PipelineBuildError]:  # noqa F821 (Hey linter, see below.)
        try:
            elements = [f() for f in self._thunks]
            logger.debug(f"_caps_string: {self._caps_string}")
            caps = self._Gst.caps_from_string(self._caps_string)
            elements[-1].set_property("caps", caps)
            pipeline = self._Gst.Pipeline()
            for x in elements:
                pipeline.add(x)
            for (x, y) in zip(elements, elements[1:]):
                # c.f. http://gstreamer-devel.966125.n4.nabble.com/Problem-linking-rtspsrc-to-any-other-element-td3051725.html
                if x.get_static_pad("src"):
                    logger.info(f"get static pad of src of {x}")
                    if not x.link(y):
                        return Err(PipelineBuildError(f"failed to link {x} {y}"))
                else:

                    def f(x: "Gst.Element", y: "Gst.Element") -> None:  # type: ignore  # noqa F821
                        logger.info(f"linking {x} and {y}")
                        x.link(y)

                    x.connect("pad-added", lambda _a, _b, x=x, y=y: f(x, y))

            return Ok(_BuiltPipeline(pipeline=pipeline, sink=elements[-1]))
        except PipelineBuildError as err:
            return Err(err)
        except Exception as err:
            try:
                raise PipelineBuildError(err) from err
            except PipelineBuildError as err:
                return Err(err)


class _BuiltPipeline(NamedTuple):
    pipeline: "Gst.Pipeline"  # type: ignore  # noqa F821
    sink: "Gst.GstAppSink"  # type: ignore  # noqa F821



import logging as _logging

# noqa idiom
if True:
    logger = _logging.getLogger(__name__)
    logger.addHandler(_logging.NullHandler())

from abc import ABC, abstractmethod
from typing import Any, Union

import PIL
from PIL.Image import Image as PIL_Image
from result import Err, Ok, Result

from ..util import _get_gst
from .pipeline import AppsinkColorFormat

__all__ = [
    "ConverterBase",
    "ConverterRaw",
    "ConverterPIL",
]


class ConverterBase(ABC):
    # Associated type.  See also `_GstStream`.
    # type ConvertResult;

    # Here, Any = Self::ConvertResult.
    @abstractmethod
    def convert_sample(
        self,
        sample: "GstSample",  # type: ignore  # noqa F821
    ) -> Result[Any, Exception]:
        """
        Convert :class:`~GstSample` to some value.
        This class is intended to be used in :class:`~_GstStream`.

        args:
            - sample: :class:`~GstSample`
        returns:
            - `Any`, depends on concrete classes
            - :class:`~Exception`
        """

        raise NotImplementedError()


class ConverterRaw(ConverterBase):
    # type ConvertResult = bytes;

    _Gst: "Gst"  # type: ignore  # noqa F821

    def __init__(self) -> None:
        self._Gst = _get_gst()

    def convert_sample(  # type: ignore  # reason: incompatible return type, but actually compatible
        self,
        sample: "GstSample",  # type: ignore  # noqa F821
    ) -> Result[bytes, RuntimeError]:
        # Note that `gst_buffer_extract_dup()` cause a memory leak.
        # c.f. https://github.com/beetbox/audioread/pull/84
        buf = sample.get_buffer()
        success, info = buf.map(self._Gst.MapFlags.READ)
        if success:
            data = info.data
            ret = bytes(data)
            buf.unmap(info)
            return Ok(ret)
        else:
            return Err(RuntimeError("`gst_buffer_map()` failed"))


class ConverterPIL(ConverterBase):
    # type ConvertResult = PIL_Image;

    _Gst: "Gst"  # type: ignore  # noqa F821

    def __init__(self) -> None:
        self._Gst = _get_gst()

    def convert_sample(  # type: ignore  # reason: incompatible return type, but actually compatible
        self,
        sample: "GstSample",  # type: ignore  # noqa F821
    ) -> Result[PIL_Image, Union[RuntimeError, ValueError]]:
        caps = sample.get_caps()
        structure = caps.get_structure(0)
        logger.debug(f"structure: {structure}")
        format_ = structure.get_value("format")
        logger.debug(f"format: {format_}")
        format__ = AppsinkColorFormat._from_caps_format(format_)
        if format__.is_err():
            return Err(format__.unwrap_err())
        raw_mode = format__.unwrap()._to_PIL_raw_mode()
        shape = (structure.get_value("width"), structure.get_value("height"))
        logger.debug(f"shape: {shape}")

        # Note that `gst_buffer_extract_dup()` cause a memory leak.
        # c.f. https://github.com/beetbox/audioread/pull/84
        buf = sample.get_buffer()
        success, info = buf.map(self._Gst.MapFlags.READ)
        if success:
            data = info.data
            ret = PIL.Image.frombytes("RGB", shape, data, "raw", raw_mode)
            buf.unmap(info)
            return Ok(ret)
        else:
            return Err(RuntimeError("`gst_buffer_map()` failed"))
